import numpy as np


def demonstration_time_series() -> (np.ndarray, np.ndarray):
    """
    Generate a time series for demonstration purposes. This is a noisy
    sine wave with one valley that is deeper than the other ones.

    Returns
    -------
    x: np.ndarray of shape (nb_samples)
        The raw time series data
    y: np.ndarray of shape (nb_samples)
        The ground truth labels
    """
    np.random.seed(42)

    X = np.sin(np.linspace(0, 25 * np.pi, 1400) + 2)
    X += np.random.normal(0, 0.1, X.shape)
    y = np.zeros(shape=X.shape[0])

    X[920:965] -= 0.5
    y[920:965] = 1

    return X, y


def inject_anomalies(
    time_series: np.ndarray,
    nb_anomalies: int = 10,
    min_anomaly_magnitude: float = 1.0,
    max_anomaly_magnitude: float = 2.0,
) -> np.array:
    """
    Inject random anomalies in the given time series. This method will
    only inject point anomalies by adding a random offset to some random
    observations in the time series. Note that this method will adapt the
    given time series in place.

    Parameters
    ----------
    time_series: array_like of shape (n_samples, n_features)
        The time series to inject anomalies in
    nb_anomalies: int, default=10
        The number of anomalies to inject
    min_anomaly_magnitude: float, default=1.0
        The minimum added magnitude for anomalies
    max_anomaly_magnitude: float, default=2.0
        The maximum added magnitude for anomalies

    Returns
    -------
    anomaly_labels: array-like of shape (n_samples)
        Binary anomaly labels, with a one indicating that an anomaly has
        been injected in the time series.
    """
    anomaly_indices = np.random.choice(
        time_series.shape[0], nb_anomalies, replace=False
    )
    if len(time_series.shape) == 1:
        anomalies = np.random.uniform(
            min_anomaly_magnitude, max_anomaly_magnitude, size=nb_anomalies
        )
        anomalies *= np.random.randint(2, size=anomalies.shape) * 2 - 1
        time_series[anomaly_indices] += anomalies
    else:
        anomalies = np.random.uniform(
            min_anomaly_magnitude,
            max_anomaly_magnitude,
            size=(nb_anomalies, time_series.shape[1]),
        )
        anomalies *= np.random.randint(2, size=anomalies.shape) * 2 - 1
        time_series[anomaly_indices, :] += anomalies
    anomaly_labels = np.zeros(shape=time_series.shape[0])
    anomaly_labels[anomaly_indices] = 1
    return anomaly_labels


def make_sine_wave(
    nb_samples: int,
    amplitude: float = 1.0,
    frequency: float = 5.0,
    phase: float = 0.0,
    noise_level: float = 0.2,
    seed: int = None,
    **kwargs,
) -> (np.ndarray, np.ndarray):
    """
    Generate a random sine wave and inject anomalies into it.

    Parameters
    ----------
    nb_samples: int
        The length of the sine wave.
    amplitude: float, default=1.0
        The amplitude of the sine wave, the max absolute value of the sine wave.
    frequency: float, default=5.0
        The frequency of the sine wave, the number of oscillations
    phase: float, default=0.0
        The phase of the sine wave, where the oscillation starts.
    noise_level: float, default=0.2
        The amount of Gaussian noise to add to the time series
    seed: int, default=None
        The seed for generating a random sine wave. If no value is provided,
        then the sine wave will be random.
    **kwargs:
        Parameters to pass to the ``inject_anomalies`` method.

    Returns
    -------
    x: np.ndarray of shape (nb_samples)
        The raw time series data
    y: np.ndarray of shape (nb_samples)
        The ground truth labels
    """
    # Generate the time series
    np.random.seed(seed)
    t = np.arange(nb_samples) / nb_samples
    nice_sine_wave = amplitude * np.sin(2 * np.pi * frequency * t + phase)
    noisy_sine_wave = nice_sine_wave + np.random.normal(0, noise_level, nb_samples)
    noisy_sine_wave = noisy_sine_wave.reshape(-1, 1)
    # Inject anomalies
    anomaly_labels = inject_anomalies(noisy_sine_wave, **kwargs)
    # Create a DataSet object
    return noisy_sine_wave, anomaly_labels



from pyod.models.hbos import HBOS

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class HistogramBasedOutlierScore(PyODAnomalyDetector):
    """
    Anomaly detector based on the Histogram Based Outlier Score (HBOS) algorithm.

    Histogram Based Outlier Score (HBOS) [goldstein2012histogram]_ constructs for each feature
    a univariate histogram. Bins with a small height (for static bin widths) or wider bins (for
    dynamic bin widths) correspond to sparse regions of the feature space. Thus, values falling
    in these bins lay in sparse regions of the feature space and are considered more anomalous.

    In this implementation, it is possible to set a window size to take the past observations into
    account. However, HBOS assumes feature independence. Therefore, for a time series with :math:`D`
    attributes and a window size :math:`w`, HBOS constructs :math:`D \\times w` independent histograms,
    from which the anomaly score is computed.

    Notes
    -----
    The HBOS detector inherets from :py:class:`~dtaianomaly.anomaly_detection.PyODAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs
        Arguments to be passed to the PyOD histogram based outlier score.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : HBOS
        An HBOS detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import HistogramBasedOutlierScore
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> hbos = HistogramBasedOutlierScore(1).fit(x)
    >>> hbos.decision_function(x)
    array([0.51808795, 0.51808795, 0.51808795, ..., 0.48347552, 0.48347552,
           0.48347552])

    References
    ----------
    .. [goldstein2012histogram] Goldstein, Markus, and Andreas Dengel. "Histogram-based outlier score (hbos):
       A fast unsupervised anomaly detection algorithm." KI-2012: poster and demo track 1 (2012): 59-63.
    """

    def _initialize_detector(self, **kwargs) -> HBOS:
        return HBOS(**kwargs)

    def _supervision(self):
        return Supervision.UNSUPERVISED



"""
This module contains preprocessing functionality.

>>> from dtaianomaly import preprocessing

Custom preprocessors can be implemented by extending the base :py:class:`~dtaianomaly.preprocessing.Preprocessor` class.
"""

from .ChainedPreprocessor import ChainedPreprocessor
from .Differencing import Differencing
from .ExponentialMovingAverage import ExponentialMovingAverage
from .MinMaxScaler import MinMaxScaler
from .MovingAverage import MovingAverage
from .PiecewiseAggregateApproximation import PiecewiseAggregateApproximation
from .Preprocessor import Identity, Preprocessor, check_preprocessing_inputs
from .RobustScaler import RobustScaler
from .StandardScaler import StandardScaler
from .UnderSampler import NbSamplesUnderSampler, SamplingRateUnderSampler

__all__ = [
    "Preprocessor",
    "check_preprocessing_inputs",
    "Identity",
    "ChainedPreprocessor",
    "MinMaxScaler",
    "StandardScaler",
    "MovingAverage",
    "ExponentialMovingAverage",
    "SamplingRateUnderSampler",
    "NbSamplesUnderSampler",
    "Differencing",
    "PiecewiseAggregateApproximation",
    "RobustScaler",
]



import abc
import inspect


class PrettyPrintable(abc.ABC):

    def __str__(self) -> str:
        return initialization_call_string(self)


def initialization_call_string(o: object) -> str:
    parameters = {
        parameter: getattr(o, parameter)
        for parameter, value in inspect.signature(o.__init__).parameters.items()
        if parameter not in ["args", "kwargs"]
        and value.default != getattr(o, parameter)
    }
    if hasattr(o, "kwargs"):
        parameters.update(o.kwargs)
    return (
        o.__class__.__name__
        + "("
        + ",".join(
            [
                f"{parameter}={string_with_apostrophe(value)}"
                for parameter, value in parameters.items()
            ]
        )
        + ")"
    )


def string_with_apostrophe(s):
    return f"'{s}'" if isinstance(s, str) else s



from typing import List, Optional, Union

import numpy as np

from dtaianomaly.anomaly_detection import BaseDetector
from dtaianomaly.preprocessing import ChainedPreprocessor, Preprocessor
from dtaianomaly.utils import is_valid_list


class Pipeline(BaseDetector):
    """
    Pipeline to combine preprocessing and anomaly detection

    The pipeline works with a single :py:class:`~dtaianomaly.preprocessing.Preprocessor` object or a
    list of :py:class:`~dtaianomaly.preprocessing.Preprocessor` objects. This list is converted into a
    :py:class:`~dtaianomaly.preprocessing.ChainedPreprocessor`. At the moment the `Pipeline` always
    requires a `Preprocessor` object passed at construction. If
    no preprocessing is desired, you need to explicitly pass an
    :py:class:`~dtaianomaly.preprocessing.Identity` preprocessor.

    Parameters
    ----------
    preprocessor: Preprocessor or list of Preprocessors
        The preprocessors to include in this pipeline.
    detector: BaseDetector
        The anomaly detector to include in this pipeline.
    """

    preprocessor: Preprocessor
    detector: BaseDetector

    def __init__(
        self,
        preprocessor: Union[Preprocessor, List[Preprocessor]],
        detector: BaseDetector,
    ):
        if not (
            isinstance(preprocessor, Preprocessor)
            or is_valid_list(preprocessor, Preprocessor)
        ):
            raise TypeError(
                "preprocessor expects a Preprocessor object or list of Preprocessors"
            )
        if not isinstance(detector, BaseDetector):
            raise TypeError("detector expects a BaseDetector object")
        super().__init__(detector.supervision)

        if isinstance(preprocessor, list):
            self.preprocessor = ChainedPreprocessor(preprocessor)
        else:
            self.preprocessor = preprocessor
        self.detector = detector

    def fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Pipeline":
        """
        Fit this pipeline to the given data.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: array-like of shape (n_samples)
            The ground truth labels, passed to the preprocessor and detector.

        Returns
        -------
        self: Pipeline
            Returns the instance itself
        """
        X, y = self.preprocessor.fit_transform(X=X, y=y)
        self.detector.fit(X=X, y=y)
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Compute raw anomaly scores.

        Note that depending on the preprocessor this output might
        not have the same shape as the input.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Raw time series

        Returns
        -------
        anomaly_scores: array-like of shape (n_samples)
            The predicted anomaly scores
        """
        X, _ = self.preprocessor.transform(X=X, y=None)
        return self.detector.decision_function(X)

    def __str__(self) -> str:
        return f"{self.preprocessor}->{self.detector}"



from typing import Dict, List, TypeVar, Union

from dtaianomaly.anomaly_detection import BaseDetector
from dtaianomaly.evaluation import BinaryMetric, Metric, ProbaMetric, ThresholdMetric
from dtaianomaly.pipeline import EvaluationPipeline
from dtaianomaly.preprocessing import Preprocessor
from dtaianomaly.thresholding import Thresholding

T = TypeVar("T")


def build_pipelines(
    preprocessors: List[Preprocessor],
    detectors: List[BaseDetector],
    metrics: List[ProbaMetric],
) -> List[EvaluationPipeline]:
    """The given lists are assumed to be non-empty."""
    return [
        EvaluationPipeline(
            preprocessor=preprocessor, detector=detector, metrics=metrics
        )
        for preprocessor in preprocessors
        for detector in detectors
    ]


def convert_to_proba_metrics(
    metrics: List[Metric], thresholds: List[Thresholding]
) -> List[ProbaMetric]:
    """The given lists are assumed to be non-empty."""
    proba_metrics = []
    for metric in metrics:
        if isinstance(metric, BinaryMetric):
            proba_metrics.extend(
                ThresholdMetric(thresholder=threshold, metric=metric)
                for threshold in thresholds
            )
        elif isinstance(metric, ProbaMetric):
            proba_metrics.append(metric)
    return proba_metrics


def convert_to_list(value: Union[T, List[T]]) -> List[T]:
    """If a list is given, it is assumed to be non-empty."""
    if not isinstance(value, list):
        return [
            value,
        ]
    return value



from typing import Optional, Tuple

import numpy as np

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class ExponentialMovingAverage(Preprocessor):
    """
    Compute exponential moving average. For a given input :math:`x`,
    the exponential moving average :math:`y` is computed as

    .. math::

       y_0 &= x_0 \\\\
       y_t &= \\alpha \\cdot x_t + (1 - \\alpha) \\cdot y_{t-1}

    with :math:`0 < \\alpha < 1` the smoothing factor. Higher values of
    :math:`\\alpha` result in more smoothing.

    Parameters
    ----------
    alpha: float
        The decaying factor to be used in the exponential moving average.
    """

    alpha: float

    def __init__(self, alpha: float) -> None:
        if not (0.0 < alpha < 1.0):
            raise ValueError("Alpha must be in the open interval ]0, 1[")
        self.alpha = alpha

    def _fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> "ExponentialMovingAverage":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        X_ = (
            np.frompyfunc(lambda a, b: self.alpha * a + (1 - self.alpha) * b, 2, 1)
            .accumulate(X)
            .astype(dtype=float)
        )
        return X_, y



import abc

import numpy as np

from dtaianomaly import utils
from dtaianomaly.PrettyPrintable import PrettyPrintable


class Thresholding(PrettyPrintable):

    @abc.abstractmethod
    def threshold(self, scores: np.ndarray) -> np.ndarray:
        """
        Apply the thresholding operation to the given anomaly scores

        Parameters
        ----------
        scores: array-like of shape (n_samples)
            The continuous anomaly scores to convert to binary anomaly labels.

        Returns
        -------
        anomaly_labels: array-like of shape (n_samples)
            The discrete anomaly labels, in which a 0 indicates normal and a
            1 indicates anomalous.
        """


class FixedCutoff(Thresholding):
    """
    Thresholding based on a fixed cut-off.

    Values higher than the cut-off are considered anomalous (1),
    values below the cut-off are considered normal (0).

    Parameters
    ----------
    cutoff: float
        The cutoff above which the given anomaly scores indicate an anomaly.
    """

    cutoff: float

    def __init__(self, cutoff: float):
        if not isinstance(cutoff, float):
            raise TypeError("Input must be a float")
        super().__init__()
        self.cutoff = cutoff

    def threshold(self, scores: np.ndarray):
        """
        Apply the cut-off thresholding.

        Parameters
        ----------
        scores: array-like (n_samples)
            Raw anomaly scores

        Returns
        -------
        anomaly_labels: array-like of shape (n_samples)
            Integer array of 1s and 0s, representing anomalous samples
            and normal samples respectively

        Raises
        ------
        ValueError
            If `scores` is not a valid array
        """
        if not utils.is_valid_array_like(scores):
            raise ValueError("Input must be numerical array-like")

        scores = np.asarray(scores)
        return np.asarray(self.cutoff <= scores, dtype=np.int8)


class ContaminationRate(Thresholding):
    """
    Thresholding based on a contamination rate.

    The top `contamination_rate` proportion of anomaly scores are considered anomalous (1),
    Other (lower) scores are considered normal (0).

    Parameters
    ----------
    contamination_rate: float
        The contamination_rate, i.e., the percentage of instances
        that are anomalous.
    """

    contamination_rate: float

    def __init__(self, contamination_rate: float):
        if not isinstance(contamination_rate, float):
            raise TypeError("Rate should be a float")
        if contamination_rate < 0.0 or 1.0 < contamination_rate:
            raise ValueError(
                f"Rate should be between 0 and 1. Received {contamination_rate}"
            )
        self.contamination_rate = contamination_rate

    def threshold(self, scores: np.ndarray):
        """
        Apply the contamination-rate thresholding.

        Parameters
        ----------
        scores: array-like (n_samples)
            Raw anomaly scores

        Returns
        -------
        anomaly_labels: array-like of shape (n_samples)
            Integer array of 1s and 0s, representing anomalous samples
            and normal samples respectively

        Raises
        ------
        ValueError
            If `scores` is not a valid array
        """
        if not utils.is_valid_array_like(scores):
            raise ValueError("Input must be numerical array-like")

        scores = np.asarray(scores)
        cutoff = np.quantile(scores, 1.0 - self.contamination_rate)
        return np.asarray(cutoff <= scores, dtype=np.int8)


class TopN(Thresholding):
    """
    Thresholding based on a top N strategy.

    The top `n` anomaly scores are considered anomalous (1),
    Other (lower) scores are considered normal (0).

    Parameters
    ----------
    n: int
        The number of instances that should be flagged as an anomaly
    """

    n: int

    def __init__(self, n: int):
        if not isinstance(n, int) or isinstance(n, bool):
            raise TypeError("Input should be an integer")
        if n < 0:
            raise ValueError(f"Expecting non-negative input. Received {n}")
        super().__init__()
        self.n = n

    def threshold(self, scores: np.ndarray):
        """
        Apply the top-N thresholding.

        Parameters
        ----------
        scores: array-like (n_samples)
            Raw anomaly scores

        Returns
        -------
        anomaly_labels: array-like of shape (n_samples)
            Integer array of 1s and 0s, representing anomalous samples
            and normal samples respectively

        Raises
        ------
        ValueError
            If `scores` is not a valid array
        """
        if not utils.is_valid_array_like(scores):
            raise ValueError("Input must be numerical array-like")
        if self.n > scores.shape[0]:
            raise ValueError(
                f"There are only {scores.shape[0]} anomaly scores given, but {self.n} should be anomalous!"
            )

        scores = np.asarray(scores)
        cutoff = np.partition(scores, -self.n)[-self.n]
        return np.asarray(cutoff <= scores, dtype=np.int8)



"""
This module contains thresholding functionality. It can be imported as follows:

>>> from dtaianomaly import thresholding

Thresholding is required to convert raw anomaly scores from a detector,
obtained via the :py:meth:`dtaianomaly.anomaly_detection.BaseDetector.decision_function`,
to binary predictions (anomaly or not).

Custom thresholders can be implemented by extending the base :py:class:`dtaianomaly.thresholding.Thresholding` class.
"""

from .thresholding import ContaminationRate, FixedCutoff, Thresholding, TopN

__all__ = ["Thresholding", "FixedCutoff", "ContaminationRate", "TopN"]



from typing import List, Optional

import numpy as np

from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision
from dtaianomaly.utils import get_dimension, is_univariate, is_valid_array_like


class DataSet:
    """
    A class for time series anomaly detection data sets. These
    consist of the raw data for training and testing anomaly
    detectors, as well as the respective ground truth labels.

    Parameters
    ----------
    X_test: array-like of shape (n_samples_test, n_attributes)
        The test time series data.
    y_test: array-like of shape (n_samples_test)
        The ground truth anomaly labels of the test data.
    X_train: array-like of shape (n_samples_train, n_attributes), optional
        The train time series. If not given, then the test data will
        be used for training and the data is only compatible with
        unsupervised anomaly detectors.
    y_train: array-like of shape (n_samples_train), optional
        The ground truth anomaly labels of the training data. If not given,
        either the train data should not be given either, or the train
        data is assumed to consist of only normal data.
    """

    X_test: np.ndarray
    y_test: np.ndarray
    X_train: np.ndarray = None
    y_train: np.ndarray = None

    def __init__(
        self,
        X_test: np.ndarray,
        y_test: np.ndarray,
        X_train: np.ndarray = None,
        y_train: np.ndarray = None,
    ):
        self.check_is_valid(X_test, y_test, X_train, y_train)
        self.X_test = X_test
        self.y_test = y_test
        self.X_train = X_train
        self.y_train = y_train

    @staticmethod
    def check_is_valid(
        X_test: np.ndarray,
        y_test: np.ndarray,
        X_train: Optional[np.ndarray],
        y_train: Optional[np.ndarray],
    ) -> None:
        """
        Checks if the given elements refer o a valid ``DataSet``. If the elements
        would not give a valid ``DataSet``, then a ``ValueError`` is raised.

        Parameters
        ----------
        X_test: array-like of shape (n_samples_test, n_attributes)
            The test time series data.
        y_test: array-like of shape (n_samples_test)
            The ground truth anomaly labels of the test data.
        X_train: array-like of shape (n_samples_train, n_attributes) or ``None``
            The train time series data. Note that, even though ``X_train`` can
            be ``None``, it must be provided.
        y_train: array-like of shape (n_samples_train) or ``None``.
            The ground truth anomaly labels of the train data. Note that, even
            though ``y_train`` can be ``None``, it must be provided.

        Raises
        ------
        ValueError:
            If the given variables would not lead to a valid ``DataSet``. This is the
            case if:

            - If ``X_test`` or ``y_test`` are not valid array-like.
            - If ``y_test`` is not univariate and has a value different from 0 or 1.
            - If ``X_test`` and ``y_test`` consist of a different number of samples.
            - If ``X_train`` is not ``None``, but it is not a valid array-like.
            - If ``X_train`` is not ``None`` and consists of a different number of
              attributes than ``X_test``.
            - If ``y_train`` is not ``None`` but ``X_train`` is ``None``.
            - If ``y_train`` is not ``None`` but it is not a valid array-like.
            - If ``y_train`` is not ``None``, but it is not univariate and has a .
              value different from 0 or 1.
            - If ``y_train`` is not ``None`` but consists of a different number of
              samples than ``X_train``.
        """
        # Check test data
        if not is_valid_array_like(X_test):
            raise ValueError("The test data must be a valid array like!")

        # Check test labels
        if not is_valid_array_like(y_test):
            raise ValueError("The test labels must be a valid array like!")
        if not is_univariate(y_test):
            raise ValueError(
                "There can only be one label for each observation in the test data!"
            )
        if not np.all(np.isin(y_test, [0, 1])):
            raise ValueError("The test labels must be binary!")
        if not y_test.shape[0] == X_test.shape[0]:
            raise ValueError(
                "The test data and labels must consist of the same number of observations!"
            )

        # Check the train data
        if X_train is not None:
            if not is_valid_array_like(X_train):
                raise ValueError("The train data must be a valid array like!")
            if get_dimension(X_test) != get_dimension(X_train):
                raise ValueError(
                    "The test and train data must consist of the same number of features!"
                )

        # Check the train data
        if y_train is not None:
            if X_train is None:
                raise ValueError(
                    "There can not be any train labels if there is no train data!"
                )
            if not is_valid_array_like(y_train):
                raise ValueError("The train labels must be a valid array like!")
            if not is_univariate(y_train):
                raise ValueError(
                    "There can only be one label for each observation in the train data!"
                )
            if not np.all(np.isin(y_train, [0, 1])):
                raise ValueError("The test labels must be binary!")
            if not X_train.shape[0] == y_train.shape[0]:
                raise ValueError(
                    "The train data and labels must consist of the same number of observations!"
                )

    def is_valid(self) -> bool:
        """
        Checks whether this ``DataSet`` is valid or not.

        Returns
        -------
        is_valid: bool
            True if and only if this instance is valid, i.e., if the attributes
            ``X_test``, ``y_test``, ``X_train`` and ``y_train`` of this instance
            pass all the checks of :py:meth:`~dtaianomaly.data.DataSet.check_is_valid`.
        """
        try:
            self.check_is_valid(
                X_test=self.X_test,
                y_test=self.y_test,
                X_train=self.X_train,
                y_train=self.y_train,
            )
            return True

        except ValueError:
            return False

    def compatible_supervision(self) -> List[Supervision]:
        """
        Get the compatible supervision types for this data set.

        Returns
        -------
        compatible_types: list of Supervision
            A list containing the compatible types for this dataset. The following
            suprvision types can be compatible:

            - ``Supervision.UNSUPERVISED``: Always compatible.
            - ``Supervision.SEMI_SUPERVISED``: Compatible if and only if there
              is some training data given (which is assumed to be normal).
            - ``Supervision.SUPERVISED``: Only compatible if both training data
              and training labels are provided.
        """
        # If there is no train data given at all, then only unsupervised detectors are compatible
        if self.X_train is None and self.y_train is None:
            return [Supervision.UNSUPERVISED]
        # If train data is given but no train labels, then either unsupervised or semi-supervised detectors are compatible
        elif self.X_train is not None and self.y_train is None:
            return [Supervision.UNSUPERVISED, Supervision.SEMI_SUPERVISED]
        # If the train data and train labels are given, then all detectors are compatible.
        else:
            return [
                Supervision.UNSUPERVISED,
                Supervision.SEMI_SUPERVISED,
                Supervision.SUPERVISED,
            ]

    def is_compatible(self, detector: BaseDetector) -> bool:
        """
        Checks if the given anomaly detector is compatible with this ``DataSet``.

        Parameters
        ----------
        detector: BaseDetector
            The anomaly detector to check if it is compatible with this ``DataSet``.

        Returns
        -------
        is_compatible: bool
            True if and only if the given anomaly detector is compatible with
            this ``DataSet``. The detector is compatible if

            - This ``DataSet`` does not contain any training data or training labels,
              only unsupervised anomaly detectors are compatible
            - This ``DataSet`` contains training data but no training labels, then
              unsupervised and semi-supervised anomaly detectors are compatible.
            - This ``DataSet`` contains training data and labels, then supervised,
              unsupervised and semi-supervised anomaly detectors are compatible.
        """
        return detector.supervision in self.compatible_supervision()



"""
This module contains functionality to detect anomalies. It can be imported
as follows:

>>> from dtaianomaly import anomaly_detection

We refer to the `documentation <https://dtaianomaly.readthedocs.io/en/stable/getting_started/anomaly_detection.html>`_
for more information regarding detecting anomalies using ``dtaianomaly``.
"""

from .BaseDetector import BaseDetector, Supervision, load_detector
from .baselines import AlwaysAnomalous, AlwaysNormal, RandomDetector
from .ClusterBasedLocalOutlierFactor import ClusterBasedLocalOutlierFactor
from .CopulaBasedOutlierDetector import CopulaBasedOutlierDetector
from .HistogramBasedOutlierScore import HistogramBasedOutlierScore
from .IsolationForest import IsolationForest
from .KernelPrincipalComponentAnalysis import KernelPrincipalComponentAnalysis
from .KMeansAnomalyDetector import KMeansAnomalyDetector
from .KNearestNeighbors import KNearestNeighbors
from .LocalOutlierFactor import LocalOutlierFactor
from .MatrixProfileDetector import MatrixProfileDetector
from .MedianMethod import MedianMethod
from .OneClassSupportVectorMachine import OneClassSupportVectorMachine
from .PrincipalComponentAnalysis import PrincipalComponentAnalysis
from .PyODAnomalyDetector import PyODAnomalyDetector
from .RobustPrincipalComponentAnalysis import RobustPrincipalComponentAnalysis
from .windowing_utils import (
    check_is_valid_window_size,
    compute_window_size,
    reverse_sliding_window,
    sliding_window,
)

__all__ = [
    # Base
    "BaseDetector",
    "Supervision",
    "load_detector",
    # Sliding window
    "sliding_window",
    "reverse_sliding_window",
    "check_is_valid_window_size",
    "compute_window_size",
    # Baselines
    "AlwaysNormal",
    "AlwaysAnomalous",
    "RandomDetector",
    # Detectors
    "ClusterBasedLocalOutlierFactor",
    "CopulaBasedOutlierDetector",
    "HistogramBasedOutlierScore",
    "IsolationForest",
    "KernelPrincipalComponentAnalysis",
    "KMeansAnomalyDetector",
    "KNearestNeighbors",
    "LocalOutlierFactor",
    "MatrixProfileDetector",
    "MedianMethod",
    "OneClassSupportVectorMachine",
    "PrincipalComponentAnalysis",
    "PyODAnomalyDetector",
    "RobustPrincipalComponentAnalysis",
]



import datetime
import os
import traceback

import dtaianomaly
from dtaianomaly.data import LazyDataLoader
from dtaianomaly.pipeline import Pipeline


def log_error(
    error_log_path: str,
    exception: Exception,
    data_loader: LazyDataLoader,
    pipeline: Pipeline = None,
    fit_on_X_train: bool = True,
) -> str:

    # Ensure the directory exists
    os.makedirs(error_log_path, exist_ok=True)

    # Set an intuitive name for the error file based on the given data loader and pipeline
    base_file_name = data_loader.__class__.__name__
    if pipeline is not None:
        base_file_name += f"-{pipeline.detector.__class__.__name__}"

    # Ensure that the file name is unique
    while True:
        now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        if not os.path.exists(f"{error_log_path}/{base_file_name}-{now}.err"):
            break
    file_path = f"{error_log_path}/{base_file_name}-{now}.err"

    # Write away the logging
    with open(file_path, "w") as error_file:

        # Create an error message as a string
        if pipeline is None:  # Didn't reach anomaly detection when error occurred.
            error_message = "An error occurred while loading data!"
        else:
            error_message = "An error occurred while detecting anomalies!"
        error_message += (
            "\nCode to reproduce the error is at the bottom of this error-log.\n\n"
        )

        # Add the error message
        error_message += "Traceback (most recent call last):\n\n"
        error_message += "\n".join(traceback.format_tb(exception.__traceback__))
        error_message += f"\n{exception.__class__.__name__}: {exception}"

        # Make sure the message is in python comments
        error_message = "# " + error_message.replace("\n", "\n# ")

        # Write the error message
        error_file.write(error_message)

        # Add an empty line below the error message
        error_file.write("\n\n")

        # Add the current version of dtaianomaly
        error_file.write("import dtaianomaly\n")
        error_file.write(
            f"assert dtaianomaly.__version__ == '{dtaianomaly.__version__}'\n"
        )

        # Ad an empty line
        error_file.write("\n")

        # Add the imports to the file
        error_file.write("from dtaianomaly.data import *\n")
        if pipeline is not None:
            error_file.write("from dtaianomaly.preprocessing import *\n")
            error_file.write("from dtaianomaly.anomaly_detection import *\n")
            error_file.write("from dtaianomaly.pipeline import Pipeline\n")

        # Ad an empty line
        error_file.write("\n")

        # Add code for loading the data
        error_file.write(f"data_loader = {data_loader}\n")
        error_file.write(f"data = data_loader.load()\n\n")

        # Add code for detecting anomalies
        if pipeline is not None:
            error_file.write(f"preprocessor = {pipeline.preprocessor}\n")
            error_file.write(f"detector = {pipeline.detector}\n")
            error_file.write(
                "pipeline = Pipeline(\n"
                "    preprocessor=preprocessor,\n"
                "    detector=detector\n"
                ")\n"
            )
            if fit_on_X_train:
                train_data = "X_train"
            else:
                train_data = "X_test"
            error_file.write(
                f"y_pred = pipeline.fit(data.{train_data}, data.y_train).predict_proba(data.X_test)\n\n"
            )

    return os.path.abspath(file_path)



from typing import Optional, Tuple

import numpy as np

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class Differencing(Preprocessor):
    """
    Applies differencing to the given time series. For a time series :math:`x`
    and given season :math:`m`, the difference :math:`y` is computed as:

    .. math::

       y_t = x_t - x_{t-m}

    This differencing process can be applied a given order of times, recursively.

    Parameters
    ----------
    order: int
        The number of times the differencing procuder should be applied. If the
        order is 0, then no differencing will be applied.
    window_size: int, default=1
        The decaying factor to be used in the exponential moving average.
    """

    order: int
    window_size: int

    def __init__(self, order: int, window_size: int = 1):
        super().__init__()

        if not isinstance(order, int) or isinstance(order, bool):
            raise TypeError("`order` should be an integer")
        if order < 0:
            raise ValueError("'order' must be positive!")

        if not isinstance(window_size, int) or isinstance(window_size, bool):
            raise TypeError("`window_size` should be an integer")
        if window_size < 1:
            raise ValueError("'window_size' must be strictly positive!")

        self.order = order
        self.window_size = window_size

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        X_ = X
        for _ in range(self.order):
            concat = np.concatenate([X_[: self.window_size], X_])
            X_ = concat[self.window_size :] - concat[: -self.window_size]
        return X_, y



from typing import Dict, List, Optional, Union

import numpy as np

from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector
from dtaianomaly.evaluation.metrics import ProbaMetric
from dtaianomaly.pipeline.Pipeline import Pipeline
from dtaianomaly.preprocessing.Preprocessor import Preprocessor
from dtaianomaly.utils import is_valid_array_like, is_valid_list


class EvaluationPipeline:
    """
    Pipeline to combine a base pipeline, and a set of metrics. Used
    in the workflow. The given :py:class:`~dtaianomaly.preprocessing.Preprocessor`
    and :py:class:`~dtaianomaly.anomaly_detection.BaseDetector` are
    combined into a :py:class:`~dtaianomaly.pipeline.Pipeline` object.

    Parameters
    ----------
    preprocessor: Preprocessor or list of Preprocessors
        The preprocessors to include in this evaluation pipeline.
    detector: BaseDetector
        The anomaly detector to include in this evaluation pipeline.
    metrics: list of Probametric objects
        The evaluation metrics to compute in this evaluation pipeline.
    """

    pipeline: Pipeline
    metrics: List[ProbaMetric]

    def __init__(
        self,
        preprocessor: Union[Preprocessor, List[Preprocessor]],
        detector: BaseDetector,
        metrics: Union[ProbaMetric, List[ProbaMetric]],
    ):
        if not (
            isinstance(metrics, ProbaMetric) or is_valid_list(metrics, ProbaMetric)
        ):
            raise TypeError("metrics should be a list of ProbaMetric objects")
        self.pipeline = Pipeline(preprocessor=preprocessor, detector=detector)
        self.metrics = metrics if isinstance(metrics, list) else [metrics]

    def run(
        self,
        X_test: np.ndarray,
        y_test: np.array,
        X_train: np.ndarray,
        y_train: Optional[np.ndarray],
    ) -> Dict[str, float]:
        """
        Run the pipeline and evaluate performance. The pipeline will
        be trained on the given train data (potentially without labels)
        and performance will be estimated on the test data.

        Parameters
        ----------
        X_test: array-like of shape (n_samples_test, n_attributes)
            The test time series data.
        y_test: array-like of shape (n_samples_test)
            The ground truth anomaly labels of the test data.
        X_train: array-like of shape (n_samples_train, n_attributes)
            The train time series data.
        y_train: array-like of shape (n_samples) or ``None``.
            The ground truth anomaly labels of the train data. Note that, even
            though ``y_train`` can be ``None``, it must be provided (i.e., there
            is no default value).

        Returns
        -------
        performances: Dict[str, float]
            The evaluation of the performance metrics. The keys are
            string descriptors of the performance metrics, with values
            the corresponding performance score.
        """
        # Validate the input
        if not is_valid_array_like(X_test):
            raise ValueError("X_test is not a valid array-like!")
        if not is_valid_array_like(y_test):
            raise ValueError("y_test is not a valid array-like!")
        if not is_valid_array_like(X_train):
            raise ValueError("X_train is not a valid array-like!")
        if not (y_train is None or is_valid_array_like(y_train)):
            raise ValueError("y_train is not a valid array-like!")

        self.fit(X_train, y_train)
        y_pred = self.predict(X_test)
        y_test_ = self.format_y_test(X_test, y_test)
        return self.evaluate(y_test_, y_pred)

    def fit(self, X_train: np.ndarray, y_train: Optional[np.array]) -> None:
        """
        Apply the fit stage of this evaluation pipeline.

        Parameters
        ----------
        X_train: array-like of shape (n_samples, n_attributes)
            The train time series data.
        y_train: array-like of shape (n_samples) or ``None``.
            The ground truth anomaly labels of the train data. Note that, even
            though ``y_train`` can be ``None``, it must be provided (i.e., there
            is no default value).
        """
        self.pipeline.fit(X_train, y_train)

    def predict(self, X_test: np.ndarray):
        """
        Apply the predict stage of the pipeline.

        Parameters
        ----------
        X_test: array-like of shape (n_samples, n_attributes)
            The test time series data.

        Returns
        -------
        y_pred: array-like of shape (n_samples)
            The predicted anomaly scores.
        """
        return self.pipeline.predict_proba(X=X_test)

    def format_y_test(self, X_test: np.ndarray, y_test: np.array) -> np.array:
        """
        Format the test labels using the preprocessor in this pipeline. This is
        necessary if some preprocessors are used that undersample the data.

        Parameters
        ----------
        X_test: array-like of shape (n_samples, n_attributes)
            The test time series data.
        y_test: array-like of shape (n_samples)
            The ground truth anomaly labels of the test data.

        Returns
        -------
        y_test_: array-like of shape (n_samples\\_)
            The formatted ground truth labels.
        """
        _, y_test_ = self.pipeline.preprocessor.transform(X_test, y_test)
        return y_test_

    def evaluate(self, y_test_: np.array, y_pred: np.array) -> Dict[str, float]:
        """
        Evaluate this pipeline by computing the evaluation scores.

        Parameters
        ----------
        y_test_: array-like of shape (n_samples)
            The formatted ground truth anomaly labels.
        y_pred: array-like of shape (n_samples)
            The predicted anomaly scores.

        Returns
        -------
        performances: Dict[str, float]
            The evaluation of the performance metrics. The keys are
            string descriptors of the performance metrics, with values
            the corresponding performance score.
        """
        return {
            str(metric): metric.compute(y_true=y_test_, y_pred=y_pred)
            for metric in self.metrics
        }



from typing import Optional, Union

import numpy as np
import stumpy
from sklearn.exceptions import NotFittedError

from dtaianomaly import utils
from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision
from dtaianomaly.anomaly_detection.windowing_utils import (
    check_is_valid_window_size,
    compute_window_size,
    reverse_sliding_window,
)


class MatrixProfileDetector(BaseDetector):
    """
    Anomaly detector based on the Matrix Profile

    Use the STOMP algorithm to detect anomalies in a time series
    [Zhu2016matrixII]_. STOMP is a fast and scalable algorithm for computing
    the matrix profile, which measures the distance from each sequence to the
    most similar other sequence. Consequently, the matrix profile can be used
    to quantify how anomalous a subsequence is, because it has a large distance
    to all other subsequences.

    Parameters
    ----------
    window_size: int or str
        The window size to use for computing the matrix profile. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    normalize : bool, default=True
        Whether to z-normalize the time series before computing
        the matrix profile.
    p : float, default=2.0
        The norm to use for computing the matrix profile.
    k : int, default=1
        The k-th nearest neighbor to use for computing the sequence distance
        in the matrix profile.
    novelty: bool, default=False
        If novelty detection should be performed, i.e., detect anomalies in regard
        to the train time series. If False, the matrix profile equals a self-join,
        otherwise the matrix profile will be computed by comparing the subsequences
        in the test data to the subsequences in the train data.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for computing the matrix profile
    X_reference_ : np.ndarray of shape (n_samples, n_attributes)
        The reference time series. Only available if ``novelty=True``

    Notes
    -----
    If the given time series is multivariate, the matrix profile is computed
    for each dimension separately and then summed up.

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import MatrixProfileDetector
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> matrix_profile = MatrixProfileDetector(window_size=50).fit(x)
    >>> matrix_profile.decision_function(x)
    array([1.20325439, 1.20690487, 1.20426043, ..., 1.47953858, 1.50188666,
           1.49891281])

    References
    ----------
    .. [Zhu2016matrixII] Y. Zhu et al., "Matrix Profile II: Exploiting a Novel
       Algorithm and GPUs to Break the One Hundred Million Barrier for Time Series
       Motifs and Joins," 2016 IEEE 16th International Conference on Data Mining
       (ICDM), Barcelona, Spain, 2016, pp. 739-748, doi: `10.1109/ICDM.2016.0085 <https://doi.org/10.1109/ICDM.2016.0085>`_.
    """

    window_size: Union[int, str]
    normalize: bool
    p: float
    k: int
    novelty: bool
    window_size_: int
    X_reference_: np.ndarray

    def __init__(
        self,
        window_size: Union[int, str],
        normalize: bool = True,
        p: float = 2.0,
        k: int = 1,
        novelty: bool = False,
    ) -> None:
        super().__init__(Supervision.UNSUPERVISED)

        check_is_valid_window_size(window_size)

        if not isinstance(normalize, bool):
            raise TypeError("`normalize` should be boolean")

        if not isinstance(p, (float, int)) or isinstance(p, bool):
            raise TypeError("`p` should be numeric")
        if p < 1.0:
            raise ValueError("`p` is a p-norm, value should be higher than 1.")

        if not isinstance(k, int) or isinstance(k, bool):
            raise TypeError("`k` should be integer")
        if k < 1:
            raise ValueError("`k` should be strictly positive")

        if not isinstance(novelty, bool):
            raise TypeError("'novelty' should be a boolean")

        self.window_size = window_size
        self.normalize = normalize
        self.p = p
        self.k = k
        self.novelty = novelty

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "MatrixProfileDetector":
        """
        Fit this detector to the given data. Function is only present for
        consistency. Only saves the given data as a numpy array if
        ``novelty=True``.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.
        kwargs:
            Additional parameters to be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.

        Returns
        -------
        self: MatrixProfileDetector
            Returns the instance itself
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")

        X = np.asarray(X)
        self.window_size_ = compute_window_size(X, self.window_size, **kwargs)
        if self.novelty:
            self.X_reference_ = np.asarray(X)

        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Compute anomaly scores.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        matrix_profile: array-like of shape (n_samples)
            Matrix profile scores. Higher is more anomalous.

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        NotFittedError
            If novelty detection must be performed (``novelty=True``), but this
            detector has not been fitted yet.
        ValueError
            If novelty detection must be performed (``novelty=True``), but the reference
            data ``X_reference_`` has a different number of attributes than the given
            data ``X``.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError(f"Input must be numerical array-like")

        # Make sure X is a numpy array
        X = np.asarray(X)

        if not hasattr(self, "window_size_"):
            raise NotFittedError("Call the fit function before making predictions!")
        if self.novelty:
            nb_attributes_test = 1 if len(X.shape) == 1 else X.shape[1]
            nb_attributes_reference = (
                1 if len(self.X_reference_.shape) == 1 else self.X_reference_.shape[1]
            )
            if nb_attributes_reference != nb_attributes_test:
                raise ValueError(
                    f"Trying to detect anomalies with Matrix Profile using ``novelty=True``, but the number of attributes "
                    f"in the reference data is different from the number of attributes in the test data: "
                    f"({nb_attributes_reference} != {nb_attributes_test})!"
                )

        # Stumpy assumes arrays of shape [C T], where C is the number of "channels"
        # and T the number of time samples

        # This function works for multivariate and univariate signals
        ignore_trivial = True if not self.novelty else False
        if len(X.shape) == 1 or X.shape[1] == 1:
            T_B = None if not self.novelty else self.X_reference_.squeeze()
            matrix_profile = stumpy.stump(
                X.squeeze(),
                T_B=T_B,
                m=self.window_size_,
                normalize=self.normalize,
                p=self.p,
                k=self.k,
                ignore_trivial=ignore_trivial,
            )[
                :, self.k - 1
            ]  # Needed if k>1?
        else:
            if self.novelty:
                matrix_profiles = np.full(
                    shape=(X.shape[0] - self.window_size_ + 1, X.shape[1]),
                    fill_value=np.nan,
                )
                for attribute in range(X.shape[1]):
                    matrix_profiles[:, attribute] = stumpy.stump(
                        X[:, attribute],
                        T_B=self.X_reference_[:, attribute],
                        m=self.window_size_,
                        normalize=self.normalize,
                        p=self.p,
                        k=self.k,
                        ignore_trivial=ignore_trivial,
                    )[:, self.k - 1]
            else:
                matrix_profiles, _ = stumpy.mstump(
                    X.transpose(),
                    m=self.window_size_,
                    discords=True,
                    normalize=self.normalize,
                    p=self.p,
                )
            matrix_profile = np.sum(matrix_profiles, axis=0)

        return reverse_sliding_window(matrix_profile, self.window_size_, 1, X.shape[0])



import abc
from typing import Optional, Tuple

import numpy as np

from dtaianomaly import utils
from dtaianomaly.PrettyPrintable import PrettyPrintable


def check_preprocessing_inputs(X: np.ndarray, y: Optional[np.ndarray] = None) -> None:
    """
    Check if the given `X` and `y` arrays are valid.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_attributes)
        Raw time series
    y: array-like, default=None
        Ground-truth information

    Raises
    ------
    ValueError
        If inputs are not valid numeric arrays
    ValueError
        If inputs have a different size in the first dimension (n_samples)
    """
    if not utils.is_valid_array_like(X):
        raise ValueError("`X` is not a valid array")
    if y is not None and not utils.is_valid_array_like(y):
        raise ValueError("`y` is not  valid array")
    if y is not None:
        X = np.asarray(X)
        y = np.asarray(y)
        if X.shape[0] != y.shape[0]:
            raise ValueError("`X` and `y` have a different number of samples")


class Preprocessor(PrettyPrintable):
    """
    Base preprocessor class.
    """

    def fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        """
        First checks the inputs with :py:meth:`~dtaianomaly.preprocessing.Preprocessor.check_preprocessing_inputs`,
        and then fits this preprocessor.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Raw time series
        y: array-like, default=None
            Ground-truth information

        Returns
        -------
        self: Preprocessor
            Returns the fitted instance self.
        """
        check_preprocessing_inputs(X, y)
        return self._fit(np.asarray(X), y if y is None else np.asarray(y))

    @abc.abstractmethod
    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        """Effectively fit this preprocessor, without checking the inputs."""

    def transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        First checks the inputs with :py:meth:`~dtaianomaly.preprocessing.Preprocessor.check_preprocessing_inputs`,
        and then transforms (i.e., preprocesses) the given time series.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Raw time series
        y: array-like of shape (n_samples), default=None
            Ground-truth information

        Returns
        -------
        X_transformed: np.ndarray of shape (n_samples, n_attributes)
            Preprocessed raw time series
        y_transformed: np.ndarray of shape (n_samples)
            The transformed ground truth. If no ground truth was provided (`y=None`),
            then None will be returned as well.
        """
        check_preprocessing_inputs(X, y)
        return self._transform(np.asarray(X), y if y is None else np.asarray(y))

    @abc.abstractmethod
    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Effectively transform the given data, without checking the inputs."""

    def fit_transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        First checks the inputs with :py:meth:`~dtaianomaly.preprocessing.Preprocessor.check_preprocessing_inputs`,
        and then chains the fit and transform methods on the given data, i.e.,
        first fit this preprocessor on the given `X` and `y`, after which the
        given `X` and `y` will be transformed.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Raw time series
        y: array-like of shape (n_samples), default=None
            Ground-truth information

        Returns
        -------
        X_transformed: np.ndarray of shape (n_samples, n_attributes)
            Preprocessed raw time series
        y_transformed: np.ndarray of shape (n_samples)
            The transformed ground truth. If no ground truth was provided (`y=None`),
            then None will be returned as well.
        """
        return self.fit(X, y).transform(X, y)


class Identity(Preprocessor):
    """
    Identity preprocessor. A dummy preprocessor which does not do any processing at all.
    """

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        return X, y



from pyod.models.ocsvm import OCSVM

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class OneClassSupportVectorMachine(PyODAnomalyDetector):
    """
    Anomaly detector based on One-Class Support Vector Machines (OC-SVM).

    The OC-SVM [Scholkopf1999support]_ uses a Support Vector Machine to learn
    a boundary around the normal behavior with minimal margin. New data can
    then be identified as anomaly or not, depending on if the data falls within
    this boundary (and thus is normal) or outside the boundary (and thus is
    anomalous).

    Notes
    -----
    The OC-SVM inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to the PyOD OC-SVM

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : OCSVM
        A OCSVM-detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import OneClassSupportVectorMachine
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> ocsvm = OneClassSupportVectorMachine(10).fit(x)
    >>> ocsvm.decision_function(x)
    array([-0.7442125 , -1.57019847, -1.86868112, ..., 13.33883568,
           12.6492399 , 11.8761641 ])

    References
    ----------
    .. [Scholkopf1999support] Schölkopf, Bernhard, et al. "Support vector method
       for novelty detection." Advances in neural information processing systems 12
       (1999).
    """

    def _initialize_detector(self, **kwargs) -> OCSVM:
        return OCSVM(**kwargs)

    def _supervision(self):
        return Supervision.SEMI_SUPERVISED



import abc
import enum
import os.path
import pickle
from pathlib import Path
from typing import Optional, Union

import numpy as np
import scipy

from dtaianomaly import utils
from dtaianomaly.PrettyPrintable import PrettyPrintable
from dtaianomaly.thresholding.thresholding import ContaminationRate


class Supervision(enum.Enum):
    """
    An enum for the different supervision types for anomaly detectors.
    Valid supervision types are:

    - ``Unsupervised``: the anomaly detector does not need any labels or training data.
    - ``Semi-supervised``: The anomaly detector requires *normal* training data, but no training labels.
    - ``Supervised``: The anomaly detector requires both training data and training labels. The training data may contain anomalies.
    """

    UNSUPERVISED = 1
    SEMI_SUPERVISED = 2
    SUPERVISED = 3


class BaseDetector(PrettyPrintable):
    """
    Abstract base class for time series anomaly detection.

    This base class defines method signatures to build
    specific anomaly detectors. User-defined detectors
    can be used throughout the ``dtaianomaly`` by extending
    this base class.

    Parameters
    ----------
    supervision: Supervision
        The type of supervision this anomaly detector requires.
    """

    supervision: Supervision

    def __init__(self, supervision: Supervision):
        if not isinstance(supervision, Supervision):
            raise TypeError("'supervision' should be a valid 'Supervision' type!")
        self.supervision = supervision

    @abc.abstractmethod
    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "BaseDetector":
        """
        Abstract method, fit this detector to the given data.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: array-like, default=None
            Ground-truth information.

        Returns
        -------
        self: BaseDetector
            Returns the instance itself.
        """

    @abc.abstractmethod
    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Abstract method, compute anomaly scores.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        decision_scores: array-like of shape (n_samples)
            The computed anomaly scores.
        """

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        """
        Predict anomaly probabilities

        Estimate the probability of a sample of `X` being anomalous,
        based on the anomaly scores obtained from `decision_function`
        by rescaling them to the range of [0, 1] via min-max scaling.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        anomaly_scores: array-like of shape (n_samples)
            1D array with the same length as `X`, with values
            in the interval [0, 1], in which a higher value
            implies that the instance is more likely to be
            anomalous.

        Raises
        ------
        ValueError
            If `scores` is not a valid array.
        ValueError
            If the prediction scores from 'decision_function' are constant, but not
            in the interval [0, 1], because these values can not unambiguously be
            transformed to an anomaly probability.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")

        raw_scores = self.decision_function(X)

        min_score = np.nanmin(raw_scores)
        max_score = np.nanmax(raw_scores)
        if min_score == max_score:
            if not (0.0 <= min_score <= 1.0):
                raise ValueError(
                    "The predicted anomaly scores are constant, but not in the interval [0, 1]. "
                    "It is not clear how to transform these unambiguously to anomaly-probabilities!"
                )
            return raw_scores

        else:
            return (raw_scores - min_score) / (max_score - min_score)

    def predict_confidence(
        self,
        X: np.ndarray,
        X_train: np.ndarray = None,
        contamination: float = 0.05,
        decision_scores_given: bool = False,
    ):
        """
        Predict the confidence of the anomaly scores on the test given test data.

        This method implements ExCeeD [perini2020quantifying]_ (Example-wise Confidence
        of anomaly Detectors) to estimate the confidence. ExCeed transforms the predicted
        decision scores to probability estimates using a Bayesian approach, which enables
        to assign a confidence score to each prediction which captures the uncertainty
        of the anomaly detector in that prediction.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            The test time series for which the confidence of anomaly scores
            should be predicted.
        X_train: array-like of shape (n_samples_train, n_attributes), default=None
            The training time series, which can be used as reference. If
            ``X_train=None``, the test set is used as reference set.
        contamination: float, default=0.05
            The (estimated) contamination rate for the data, i.e., the expected
            percentage of anomalies.
        decision_scores_given: bool, default=False
            Whether the given ``X`` and ``X_train`` represent time series data
            or decision scores. If ``decision_scores_given=False`` (default),
            then the given arrays are interpreted as time series. Otherwise,
            they are interpreted as decision scores, as computed by
            ``decision_function()``.

        Returns
        -------
        confidence: array-like of shape (n_samples)
            The confidence of this anomaly detector in each prediction in the
            given test time series.

        References
        ----------
        .. [perini2020quantifying] Perini, L., Vercruyssen, V., Davis, J. Quantifying
           the Confidence of Anomaly Detectors in Their Example-Wise Predictions. In:
           Machine Learning and Knowledge Discovery in Databases. ECML PKDD 2020.
           Springer, Cham, doi: `10.1007/978-3-030-67664-3_14 <https://doi.org/10.1007/978-3-030-67664-3_14>`_.
        """
        # Set the decision scores
        if decision_scores_given:
            if len(X.shape) > 1:
                raise ValueError(
                    "In the 'predict_confidence()' method, it was indicated that the decision scores are provided "
                    "as X (decision_scores_given=True), but the shape of X does not correspond to the shape of decision"
                    f"scores: {X.shape}!"
                )
            if X_train is not None and len(X_train.shape) > 1:
                raise ValueError(
                    "In the 'predict_confidence()' method, it was indicated that the decision scores are provided "
                    "as X (decision_scores_given=True), but the shape of X_train does not correspond to the shape of decision"
                    f"scores: {X.shape}!"
                )
            decision_scores = X
            decision_scores_train = X_train if X_train is not None else decision_scores

        else:
            # Compute the decision scores
            decision_scores = self.decision_function(X)
            decision_scores_train = (
                self.decision_function(X_train)
                if X_train is not None
                else decision_scores
            )

        # Convert the decision scores to binary predictions
        prediction = ContaminationRate(contamination_rate=contamination).threshold(
            decision_scores
        )

        # Apply the ExCeed method (https://github.com/Lorenzo-Perini/Confidence_AD/blob/master/ExCeeD.py)
        n = decision_scores.shape[0]

        count_instances = np.vectorize(
            lambda x: np.count_nonzero(decision_scores_train <= x)
        )
        n_instances = count_instances(decision_scores)

        prob_func = np.vectorize(lambda x: (1 + x) / (2 + n))
        posterior_prob = prob_func(
            n_instances
        )  # Outlier probability according to ExCeeD

        conf_func = np.vectorize(
            lambda p: 1 - scipy.stats.binom.cdf(n - int(n * contamination), n, p)
        )
        exWise_conf = conf_func(posterior_prob)
        np.place(
            exWise_conf, prediction == 0, 1 - exWise_conf[prediction == 0]
        )  # if the example is classified as normal, use 1 - confidence.

        return exWise_conf

    def save(self, path: Union[str, Path]) -> None:
        """
        Save detector to disk as a pickle file with extension `.dtai`. If the given
        path consists of multiple subdirectories, then the not existing subdirectories
        are created.

        Parameters
        ----------
        path: str or Path
            Location where to store the detector.
        """
        # Add the '.dtai' extension
        if Path(path).suffix != ".dtai":
            path = f"{path}.dtai"

        # Create the subdirectory, if it doesn't exist
        if not os.path.exists(Path(path).parent):
            os.makedirs(Path(path).parent)

        # Effectively write the anomaly detector to disk
        with open(path, "wb") as f:
            pickle.dump(self, f)


def load_detector(path: Union[str, Path]) -> BaseDetector:
    """
    Load a detector from disk.

    Warning: method relies on pickle. Only load trusted files!

    Parameters
    ----------
    path: str or Path
        Location of the stored detector.

    Returns
    -------
    detector: BaseDetector
        The loaded detector.
    """
    with open(path, "rb") as f:
        detector = pickle.load(f)
    return detector
